'''
Metrics to measure forecast error

ME - mean error
MAE - mean absolute error
MSE - mean squared error
RMSE - root mean squared error
MAPE - mean absolute percentage error
sMAPE - symmetric MAPE.
MASE - mean absolute scaled error

coverage - prediction interval coverage
'''
import numpy as np

from forecast_tools.baseline import SNaive


def as_arrays(y_true, y_pred):
    '''
    Returns ground truth and predict
    values as numpy arrays.

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- array-like
        the predictions

    Returns:
    -------
    Tuple(np.array np.array)
    '''
    return np.asarray(y_true), np.asarray(y_pred)


def mean_error(y_true, y_pred):
    '''
    Computes Mean Error (ME).

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- arraylike
        the predictions to evaluate

    Returns:
    -------
    float
        scalar value representing the ME
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)
    return np.mean(y_true - y_pred)


def mean_absolute_percentage_error(y_true, y_pred):
    '''
    Mean Absolute Percentage Error (MAPE).

    MAPE is a relative error measure of forecast accuracy.

    Limitations of MAPE ->

    1. When the ground true value is close to zero MAPE is inflated.

    2. MAPE is not symmetric.  MAPE produces smaller forecast
    errors when underforecasting.

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- arraylike
        the predictions to evaluate

    Returns:
    -------
    float,
        scalar value representing the MAPE (0-100)
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)
    return np.mean(np.abs((y_true - y_pred) / y_true)) * 100


def mean_absolute_error(y_true, y_pred):
    '''
    Mean Absolute Error (MAE)

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- arraylike
        the predictions to evaluate

    Returns:
    -------
    float,
        scalar value representing the MAE
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)
    return np.mean(np.abs((y_true - y_pred)))


def mean_squared_error(y_true, y_pred):
    '''
    Mean Squared Error (MSE)

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- arraylike
        the predictions to evaluate

    Returns:
    -------
    float,
        scalar value representing the MSE
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)
    return np.mean(np.square((y_true - y_pred)))


def root_mean_squared_error(y_true, y_pred):
    '''
    Root Mean Squared Error (RMSE).

    Square root of the mean squared error.

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- arraylike
        the predictions to evaluate

    Returns:
    -------
    float,
        scalar value representing the RMSE
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)
    return np.sqrt(mean_squared_error(y_true, y_pred))


def symmetric_mean_absolute_percentage_error(y_true, y_pred):
    '''
    Symmetric Mean Absolute Percentage Error (sMAPE)

    A proposed improvement./replacement for MAPE.  (But still not symmetric).

    Computation based on Hyndsight blog:
    https://robjhyndman.com/hyndsight/smape/

    Limitations of sMAPE:

    1. When the ground true value is close to zero MAPE is inflated.
    2. Like MAPE it is not symmetric.

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series
    y_pred -- arraylike
        the predictions to evaluate

    Returns:
    -------
    float,
        scalar value representing the RMSE
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)
    numerator = 2 * np.abs(y_true - y_pred)
    denominator = np.abs(y_pred) + np.abs(y_true)
    return np.mean(100 * (numerator / denominator))


def coverage(y_true, pred_intervals):
    '''
    Prediction Interval Coverage

    Calculates the proportion of the true
    values are that are covered by the lower
    and upper bounds of the prediction intervals

    Parameters:
    -------
    y_true -- array-like,
        actual observations

    pred_intervals -- np.array, matrix (hx2)
        prediction intervals

    Returns:
    -------
    float
    '''
    y_true = np.asarray(y_true)
    lower = np.asarray(pred_intervals.T[0])
    upper = np.asarray(pred_intervals.T[1])

    cover = len(np.where((y_true > lower) & (y_true < upper))[0])
    return cover / len(y_true)


def mean_absolute_scaled_error(y_true, y_pred, y_train, period=None):
    '''
    Mean absolute scaled error (MASE)

    MASE = MAE / MAE_{insample, naive}

    For definition: https://otexts.com/fpp2/accuracy.html

    Parameters:
    --------
    y_true: array-like
        actual observations from time series

    y_pred: array-like
        the predictions to evaluate

    y_train: array-like
        the training data the produced the predictions

    period: int or None, optional (default = None)
        if None then out of sample MAE is scaled by 1-step in-sample Naive1
        MAE.  If = int then SNaive is used as the scaler.

    Returns:
    -------
    float,
        scalar value representing the MASE
    '''
    y_true, y_pred = as_arrays(y_true, y_pred)

    if period is None:
        period = 1

    in_sample = SNaive(period=period)
    in_sample.fit(y_train)

    mae_insample = mean_absolute_error(y_train[period:],
                                       in_sample.fittedvalues.dropna())

    return mean_absolute_error(y_true, y_pred) / mae_insample


def forecast_errors(y_true, y_pred, metrics='all'):
    '''
    Convenience function for return a multiple
    forecast errors

    Parameters:
    --------
    y_true -- array-like
        actual observations from time series

    y_pred -- array-like
        the predictions to evaluate

    metrics -- str or List
        forecast error metrics to compute.
        'all' returns all forecast errors available
        List options: ['me', 'mae', 'mse', 'rmse', 'mape', 'smape']

    Returns:
    -------
    dict,
        forecast error metrics

    Example:
    ---------
    >>> y_true = [45, 60, 23, 45]
    >>> y_preds = [50, 50, 50, 50]

    >>> metrics = forecast_errors(y_true, y_preds)
    >>> print(metrics)

    >>> metrics = forecast_errors(y_true, y_preds, metrics=['mape', 'smape'])
    >>> print(metrics)

    '''
    y_true, y_pred = as_arrays(y_true, y_pred)

    if metrics == 'all':
        metrics = ['me', 'mae', 'mse', 'rmse', 'mape', 'smape']

    funcs = _forecast_error_functions()
    errors = {}
    for metric in metrics:
        errors[metric] = funcs[metric](y_true, y_pred)

    return errors


def _forecast_error_functions():
    '''
    Return all forecast functions in
    a dict

    Returns:
    --------
        dict
    '''
    funcs = {}
    funcs['me'] = mean_error
    funcs['mae'] = mean_absolute_error
    funcs['mse'] = mean_squared_error
    funcs['rmse'] = root_mean_squared_error
    funcs['mape'] = mean_absolute_percentage_error
    funcs['smape'] = symmetric_mean_absolute_percentage_error
    return funcs


if __name__ == '__main__':
    y_true = [45, 60, 23, 45]
    y_preds = [50, 50, 50, 50]

    metrics = forecast_errors(y_true, y_preds)
    print(metrics)

    metrics = forecast_errors(y_true, y_preds, metrics=['mape', 'smape'])
    print(metrics)



'''
Contains functions for loading built-in datasets
'''

import pandas as pd
import os

PATH_ED = 'data/ed_ts.csv'


def load_emergency_dept():
    '''
    344 simulated daily level attendenances at a single
    emergency department between 22/01/2017
    and 31/02/2017 (dates are UK dd/mm/yyyy format)

    The returned data frame has a DateTimeIndex with
    freq 'D' and value column 'arrivals' shape = (344, 1)

    Returns
    -------
        pandas.DataFrame
    '''
    dir_path = os.path.dirname(os.path.realpath(__file__))
    path = os.path.join(dir_path, PATH_ED)
    df = pd.read_csv(path, index_col='date', parse_dates=True,
                     dayfirst=True)
    df.index.freq = 'D'
    return df



'''
forecast_tools.feature_engineering

Utilities to support the creation of features for forecasting

'''

import numpy as np

def sliding_window(train, window_size=2, horizon=1):
    '''
    Time series sliding window 

    Transforms a univariate time series into a 
    supervised learning problem with features and
    target vectors. Features are lagged observations
    in the series with lag_max = window_size.  The 
    target vector is of size horizon.
    
    Parameters:
    -----------

    train: array-like
        training data for time series method
    
    window_size: int, optional (default=2)
        lookback - the maximum lag to include in the
        features.
        
    horizon: int, optional (default=1)
        number of observations ahead to predict
            
    Returns:
    -------
        array-like, array-like
    
        preprocessed X, preprocessed Y
    '''
    tabular_X = []
    tabular_y = []
    
    for i in range(0, len(train) - window_size - horizon):
        X_train = train[i:window_size+i]
        y_train = train[i+window_size+horizon-1]
        tabular_X.append(X_train)
        tabular_y.append(y_train)
       
    return np.asarray(tabular_X), np.asarray(tabular_y).reshape(-1, 1)


import pytest

from forecast_tools import datasets as ds


def test_load_ed_shape():
    expected_shape = (344, 1)
    df = ds.load_emergency_dept()
    assert df.shape == expected_shape


def test_load_ed_freq():
    expected_freq = 'D'
    df = ds.load_emergency_dept()
    assert df.index.freq.freqstr == expected_freq



'''
model_selection module:

Provides:
Tools to support the selection of the best forecasting model

In particular tools that support cross validation.

1. rolling_forecast_origin
    A train-test split generator.  Starting from a specified origin the
    generator returns (train, test) splits where the origin of test
    incrementally moves forward in time.

2. sliding_window
    A train-test split generator.  Similar to rolling_forecast_origin.
    However there is a 'window size' i.e. the training set is a fixed size.

3. cross_validation_score
    Uses a train, test split generator, a model and a forecast error metric to
    return statistics on a models prediction performance.

Dev notes:
--------

At the moment coverage is not included.  I have handled this in practical
projects by including a function that returns all of the predictions within a
split/fold. A seperate function then handles coverage.
'''

import numpy as np
from joblib import Parallel, delayed

from forecast_tools.metrics import (mean_absolute_scaled_error,
                                    _forecast_error_functions)
from forecast_tools.baseline import baseline_estimators


def rolling_forecast_origin(train, min_train_size, horizon, step=1):
    '''
    Rolling forecast origin generator.

    Parameters:
    --------
    train: array-like
        training data for time series method

    min_train_size: int
        lookback - initial training size

    horizon: int
        forecast horizon.

    step: int, optional (default=1)
        step=1 means that a single additional data point is added to the time
        series.  increase step to run less splits.

    Returns:
        array-like, array-like

        split_training, split_validation, horizons
    '''

    for i in range(0, len(train) - min_train_size - horizon + 1, step):
        split_train = train[:min_train_size+i]
        split_val = train[min_train_size+i:min_train_size+i+horizon]
        yield split_train, split_val


def sliding_window(train, window_size, horizon, step=1):
    '''
    sliding window  generator.

    Parameters:
    --------
    train: array-like
        training data for time series method

    window_size: int
        lookback - how much data to include.

    horizon: int
        forecast horizon.

    step: int, optional (default=1)
        step=1 means that a single additional data point is added to the time
        series.  increase step to run less splits.

    Returns:
        array-like, array-like

        split_training, split_validation, horizons
    '''

    for i in range(0, len(train) - window_size - horizon + 1, step):
        split_train = train[i:window_size+i]
        split_val = train[i+window_size:window_size+i+horizon]
        yield split_train, split_val


def forecast_accuracy(model, train, test, horizons, metric):
    '''
    Forecast accuracy of a model over multiple horizons

    Returns forecast accuracy of model fit on
    training data and compared against a test set
    with a given metric.

    Allows multiple forecast horizons.  The model predicts
    the maximum forecast horizon and then calculates the
    accuracy across each.

    Parameters:
    ----------
    model - object
        forecasting model with .fit(train) and
        .predict(horizon) methods

    train - array-like
        training data

    test: array-like
        holdout data for testing

    horizons: list
        list of forecast horizons e.g. [7, 14, 28]

    metric: function
        error measure sig (y_true, y_preds)

    Returns:
    -------
    float
    '''
    h_accuracy = []
    model.fit(train)
    preds = model.predict(max(horizons))
    for horizon in horizons:
        score = metric(y_true=test[:horizon], y_pred=preds[:horizon])
        h_accuracy.append(score)
    return h_accuracy


def cross_validation_score(model, cv, metric, horizons=None,
                           n_jobs=-1):
    '''
    Cross validation scores

    Parameters:
    ----------
    model: object
        forecast model

    metric: func(y_true, y_pred)
        forecast error metric

    horizons: list, optional (default=None):
        If the user wishes to return cross validation results
        for multiple sub-horizons e.g. within a 28 horizon [7, 14, 28]

    n_jobs: int, optional (default=-1)
        when -1 runs across all cores
        set = 1 to run each cross validation seperately.
        using -1 speeds up cross validation of slow running models.

    Returns:
    -------
    array of arrays
    '''

    if horizons is None:
        cv_scores = \
            Parallel(n_jobs=n_jobs)(delayed(forecast_accuracy)(model,
                                                               cv_train,
                                                               cv_test,
                                                               [len(cv_test)],
                                                               metric)
                                    for cv_train, cv_test in cv)
    else:

        cv_scores = \
            Parallel(n_jobs=n_jobs)(delayed(forecast_accuracy)(model,
                                                               cv_train,
                                                               cv_test,
                                                               horizons,
                                                               metric)
                                    for cv_train, cv_test in cv)

    return np.array(cv_scores)


def forecast(model, train, test, horizon):
    '''
    h-step prediction of a model

    Returns a tuple of (y_preds, y_train, y_true) of model fit
    to training data

    Parameters:
    ----------
    model - object
        forecasting model with .fit(train) and
        .predict(horizon) methods

    train - array-like
        training data

    test: array-like
        holdout data for testing

    horizon: int
        forecast horizon

    Returns:
    --------
    tuple (y_pred, y_train, y_true)
    '''
    y_pred = model.fit_predict(train, horizon)
    return train, test, y_pred


def cross_validation_folds(model, cv, n_jobs=-1):
    '''
    Cross validation forecasts

    Parameters:
    ----------
    model: object
        forecast model

    cv: object
        cross validation generator
        i.e. rolling_forecast_origin or sliding_window

    n_jobs: int, optional (default=-1)
        when -1 runs across all cores
        set = 1 to run each cross validation seperately.
        using -1 speeds up cross validation of slow running models.

    Returns:
    -------
    np.ndarray of tuples
    each tuple is (cv_train, cv_test, cv_y_pred)
    '''

    cv_folds = \
        Parallel(n_jobs=n_jobs)(delayed(forecast)(model,
                                                  cv_train,
                                                  cv_test,
                                                  len(cv_test))
                                for cv_train, cv_test in cv)

    return np.array(cv_folds)


def scaled_cross_validation_score(model, cv, seasonal_period=None):
    '''
    Mean absolute scaled error cross validation score

    Parameters:
    ----------
    model: object
        forecast model

    cv: generator
        time series cross validation fold generator

    metric: func(y_true, y_py_red)
        forecast error metric

    seasonal_period: None or int, optional (default=None)
        if none the in-sample one step Naive1 used for scaling.
        if int SNaive is used instead.

    Returns:
    --------
    np.array of mase scores
    '''

    folds = cross_validation_folds(model, cv)

    scores = []
    for y_train, y_true, y_pred in folds:
        score = mean_absolute_scaled_error(y_true, y_pred, y_train,
                                           period=seasonal_period)
        scores.append(score)

    return np.array(scores)


def auto_naive(y_train, horizon=1, seasonal_period=1, min_train_size='auto',
               method='cv', step=1, window_size='auto', metric='mae'):
    '''Automatic selection of the 'best' naive benchmark on a 'single' series

    The selection process uses out-of-sample cv performance.

    By default auto_naive uses cross validation to estimate the mean
    point forecast peformance of all naive methods.  It selects the method
    with the lowest point forecast metric on average.

    If there is limited data for training a basic holdout sample could be
    used.

    Dev note: the plan is to update this to work with multiple series.
    It would be best to use MASE for multiple series comparison.

    Parameters:
    ----------
    y_train: array-like
        training data.  typically in a pandas.Series, pandas.DataFrame
        or numpy.ndarray format.

    horizon: int, optional (default=1)
        Forecast horizon.

    seasonal_period: int, optional (default=1)
        Frequency of the data.  E.g. 7 for weekly pattern, 12 for monthly
        365 for daily.

    min_train_size: int or str, optional (default='auto')
        The size of the initial training set (if method=='ro' or 'sw').
        If 'auto' then then min_train_size is set to len(y_train) // 3
        If main_train_size='auto' and method='holdout' then
        min_train_size = len(y_train) - horizon.

    method: str, optional (default='cv')
        out of sample selection method.
        'ro' - rolling forecast origin
        'sw' - sliding window
        'cv' - scores from both ro and sw
        'holdout' - single train/test split
         Methods'ro' and 'sw' are similar, however, sw has a fixed
         window_size and drops older data from training.

    step: int, optional (default=1)
        The stride/step of the cross-validation. I.e. the number
        of observations to move forward between folds.

    window_size: str or int, optional (default='auto')
        The window_size if using sliding window cross validation
        When 'auto' and method='sw' then
        window_size=len(y_train) // 3

    metric: str, optional (default='mae')
        The metric to measure out of sample accuracy.
        Options: mase, mae, mape, smape, mse, rmse, me.

    Returns:
    --------
    dict
        'model': baseline.Forecast
        f'{metric}': float

        Contains the model and its CV performance.

    Raises:
    -------
    ValueError
        For invalid method, metric, window_size parameters

    See Also:
    --------
    forecast_tools.baseline.Naive1
    forecast_tools.baseline.SNaive
    forecast_tools.baseline.Drift
    forecast_tools.baseline.Average
    forecast_tools.baseline.EnsembleNaive
    forecast_tools.baseline.baseline_estimators
    forecast_tools.model_selection.rolling_forecast_origin
    forecast_tools.model_selection.sliding_window
    forecast_tools.model_selection.mase_cross_validation_score
    forecast_tools.metrics.mean_absolute_scaled_error

    Examples:
    ---------
    Measuring MAE and taking the best method using both
    rolling origin and sliding window cross validation
    of a 56 day forecast.

    >>> from forecast_tools.datasets import load_emergency_dept
    >>> y_train = load_emergency_dept
    >>> best = auto_naive(y_train, seasonal_period=7, horizon=56)
    >>> best
    {'model': Average(), 'mae': 19.63791579700355}


    Take a step of 7 days between cv folds.

    >>> from forecast_tools.datasets import load_emergency_dept
    >>> y_train = load_emergency_dept
    >>> best = auto_naive(y_train, seasonal_period=7, horizon=56,
        ...               step=7)
    >>> best
    {'model': Average(), 'mae': 19.675635558539383}

    '''
    valid_methods = ['holdout', 'ro', 'sw', 'cv']
    metrics = _forecast_error_functions()
    metrics['mase'] = mean_absolute_scaled_error

    if method not in valid_methods:
        raise ValueError(f"Method must be in {valid_methods}")

    if metric not in metrics:
        raise ValueError(f"Please select a metric from {list(metrics.keys())}")

    if min_train_size == 'auto':
        min_train_size = len(y_train) // 3
    elif not type(min_train_size) is int:
        raise ValueError(f"valid min_train_size values are 'auto' or int > 0")
    elif min_train_size < 1:
        raise ValueError(f"valid min_train_size values are 'auto' or int > 0")

    if window_size == 'auto':
        window_size = len(y_train) // 3
    elif not type(window_size) is int:
        raise ValueError(f"valid window_size values are 'auto' or int > 0")
    elif window_size < 1:
        raise ValueError(f"valid window_size values are 'auto' or int > 0")

    baselines = baseline_estimators(seasonal_period)

    method_score = []
    if method == 'cv':
        for _, model in baselines.items():
            cv_ro = rolling_forecast_origin(train=y_train,
                                            min_train_size=min_train_size,
                                            horizon=horizon,
                                            step=step)

            cv_sw = sliding_window(train=y_train,
                                   window_size=window_size,
                                   horizon=horizon,
                                   step=step)

            if metric == 'mase':
                score_ro = scaled_cross_validation_score(model, cv_ro,
                                                         seasonal_period)
                score_sw = scaled_cross_validation_score(model, cv_sw,
                                                         seasonal_period)

            else:
                score_ro = cross_validation_score(model, cv_ro,
                                                  metrics[metric])
                score_sw = cross_validation_score(model, cv_sw,
                                                  metrics[metric])

            score = np.concatenate([score_ro, score_sw])
            method_score.append(score.mean())

    elif method == 'ro':
        for _, model in baselines.items():
            cv = rolling_forecast_origin(train=y_train,
                                         min_train_size=min_train_size,
                                         horizon=horizon,
                                         step=step)

            if metric == 'mase':
                score_ro = scaled_cross_validation_score(model, cv,
                                                         seasonal_period)

            else:
                score_ro = cross_validation_score(model, cv, metrics[metric])

            method_score.append(score_ro.mean())

    elif method == 'sw':
        for _, model in baselines.items():
            cv = sliding_window(train=y_train,
                                window_size=window_size,
                                horizon=horizon,
                                step=step)

            if metric == 'mase':
                score_sw = scaled_cross_validation_score(model, cv,
                                                         seasonal_period)

            else:
                score_sw = cross_validation_score(model, cv, metrics[metric])

            method_score.append(score_sw.mean())

    else:
        # single train test split
        min_train_size = len(y_train) - horizon
        train = y_train[:min_train_size]
        test = y_train[min_train_size:]

        for _, model in baselines.items():
            model.fit(train)
            y_preds = model.predict(horizon)
            if metric == 'mase':
                score = metrics[metric](test, y_preds, y_train,
                                        seasonal_period)
            else:
                score = metrics[metric](test, y_preds)

            method_score.append(score.mean())

    method_score = np.array(method_score)
    best_index = np.argmin(method_score)

    best = {'model': list(baselines.items())[best_index][1],
            f'{metric}': method_score[best_index]}

    return best



'''
Tests for model selection / cross validation

At the moment these are all positive test.
This needs to be updated to include failure/exception testing.

'''


import pytest
import numpy as np
import pandas as pd

from forecast_tools import model_selection as ms
from forecast_tools import metrics
from forecast_tools import baseline as b


@pytest.mark.parametrize("train_size, horizon, expected",
                         [(10, 1, 1),
                          (100, 1, 1),
                          (100, 10, 10),
                          (18, 6, 6),
                          (12, 3, 3)])
def test_rfo_test_length(train_size, horizon, expected):
    '''
    test the length of test is correct in the rolling origin
    generator
    '''
    train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(train, min_train_size=1, horizon=horizon)

    _, test_cv = next(cv)

    assert expected == len(test_cv)


@pytest.mark.parametrize("train_size, horizon, expected",
                         [(10, 1, 1),
                          (100, 1, 1),
                          (100, 10, 10),
                          (18, 6, 6),
                          (12, 3, 3)])
def test_sw_test_length(train_size, horizon, expected):
    '''
    test the length of test is correct in the sliding window
    generator
    '''
    train = np.arange(train_size)
    cv = ms.sliding_window(train, window_size=1, horizon=horizon)

    _, test_cv = next(cv)

    assert expected == len(test_cv)


@pytest.mark.parametrize("train_size, min_train_size, horizon, expected",
                         [(10, 3, 1, 3),
                          (100, 34, 1, 34),
                          (100, 84, 10, 84),
                          (18, 12, 6, 12),
                          (12, 1, 3, 1)])
def test_rfo_min_train_length(train_size, min_train_size, horizon, expected):
    '''
    check that the minimum training size is correct
    '''
    train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(train, min_train_size=min_train_size,
                                    horizon=horizon)

    train_cv, _ = next(cv)

    assert expected == len(train_cv)


@pytest.mark.parametrize("train_size, window_size, horizon, expected",
                         [(10, 3, 1, 3),
                          (100, 34, 1, 34),
                          (100, 84, 10, 84),
                          (18, 12, 6, 12),
                          (12, 1, 3, 1)])
def test_sw_first_window_size(train_size, window_size, horizon, expected):
    '''
    check that the window size is correct in the sliding
    window method - this tests the first fold.
    '''
    train = np.arange(train_size)
    cv = ms.sliding_window(train, window_size=window_size,
                           horizon=horizon)

    train_cv, _ = next(cv)

    assert expected == len(train_cv)


@pytest.mark.parametrize("train_size, window_size, horizon, expected",
                         [(10, 3, 1, 3),
                          (100, 34, 1, 34),
                          (100, 84, 10, 84),
                          (18, 6, 6, 6),
                          (12, 1, 3, 1)])
def test_sw_second_window_size(train_size, window_size, horizon, expected):
    '''
    check that the window size is correct in the sliding
    window method - this tests the second fold.
    '''
    train = np.arange(train_size)
    cv = ms.sliding_window(train, window_size=window_size,
                           horizon=horizon)

    train_cv, _ = next(cv)
    # test the second fold
    train_cv, _ = next(cv)

    assert expected == len(train_cv)


@pytest.mark.parametrize("train_size, min_train_size, horizon, step, expected",
                         [(10, 3, 1, 1, 4),
                          (100, 34, 1, 10, 44),
                          (100, 3, 10, 9, 12),
                          (10000, 999, 6, 34, 999+34),
                          (12, 1, 3, 5, 6)])
def test_rfo_second_fold_train_size(train_size, min_train_size, horizon, step,
                                    expected):
    '''
    check that the second fold size = minimum training size + step
    '''
    train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(train, min_train_size=min_train_size,
                                    horizon=horizon, step=step)

    train_cv, _ = next(cv)
    # second fold min_train_size + step == len
    train_cv, _ = next(cv)

    assert expected == len(train_cv)


@pytest.mark.parametrize("n_horizons, expected",
                         [(1, 1),
                          (2, 2),
                          (3, 3),
                          (15, 15),
                          (45, 45)])
def test_forecast_accuracy_length(n_horizons, expected):
    model = b.Naive1()
    train = np.arange(10000)
    metric = metrics.mean_absolute_error
    horizons = np.arange(1, n_horizons+1).tolist()
    cv = ms.rolling_forecast_origin(train, min_train_size=100,
                                    horizon=100)
    train_cv, test_cv = next(cv)

    result_h = ms.forecast_accuracy(model, train_cv, test_cv,
                                    horizons=horizons, metric=metric)

    assert len(result_h) == expected


@pytest.mark.parametrize("train_size, min_train_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_rfo_number_of_folds(train_size, min_train_size, horizon, step,
                             expected):
    '''
    check that the number of folds returned from rolling origin
    is as expected
    '''
    train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(train, min_train_size=min_train_size,
                                    horizon=horizon, step=step)
    actual = 0

    # number of folds found
    for _, _ in cv:
        actual += 1

    assert expected == actual


@pytest.mark.parametrize("train_size, window_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_sw_number_of_folds(train_size, window_size, horizon, step,
                            expected):
    '''
    check that the number of folds returned from sliding window
    is as expected
    '''
    train = np.arange(train_size)
    cv = ms.sliding_window(train, window_size=window_size,
                           horizon=horizon, step=step)
    actual = 0

    # number of folds found
    for _, _ in cv:
        actual += 1

    assert expected == actual


@pytest.mark.parametrize("train_size, window_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_sw_number_of_folds_pd(train_size, window_size, horizon, step,
                               expected):
    '''
    check that the number of folds returned from rolling origin
    is as expected when data source is a PANDAS.DATAFRAME
    '''
    train = pd.DataFrame(np.arange(train_size))
    cv = ms.sliding_window(train, window_size=window_size,
                           horizon=horizon, step=step)
    actual = 0

    # number of folds found
    for _, _ in cv:
        actual += 1

    assert expected == actual


@pytest.mark.parametrize("train_size, min_train_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_rfo_number_of_folds_pd(train_size, min_train_size, horizon, step,
                                expected):
    '''
    check that the number of folds returned from rolling origin
    is as expected when data source is a pandas.DataFrame
    '''
    train = pd.DataFrame(np.arange(train_size))
    cv = ms.rolling_forecast_origin(train, min_train_size=min_train_size,
                                    horizon=horizon, step=step)
    actual = 0

    # number of folds found
    for _, _ in cv:
        actual += 1

    assert expected == actual


@pytest.mark.parametrize("train_size, min_train_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_mase_cv_number_of_folds(train_size, min_train_size, horizon, step,
                                 expected):
    '''
    check that the number of folds returned from rolling origin
    is as expected when data source is a pandas.DataFrame
    '''
    train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(train, min_train_size=min_train_size,
                                    horizon=horizon, step=step)

    scores = ms.scaled_cross_validation_score(b.Naive1(), cv)

    print(expected, len(scores))
    print(scores)
    assert expected == len(scores)


@pytest.mark.parametrize("train, metric",
                         [(np.arange(7), 'mae'),
                          (np.arange(7), 'mase')])
def test_auto_naive_return_length(train, metric):
    result = ms.auto_naive(train, metric=metric)
    assert len(result) == 2


@pytest.mark.parametrize("train, cv, metric",
                         [(np.arange(7), 'ro', 'mase'),
                          (np.arange(7), 'sw', 'mase'),
                          (np.arange(7), 'holdout', 'mase'),
                          (np.arange(7), 'ro', 'mae'),
                          (np.arange(7), 'sw', 'mae'),
                          (np.arange(7), 'holdout', 'mae')])
def test_auto_naive_return_length_vary_cv(train, cv, metric):
    result = ms.auto_naive(train, method=cv, metric=metric)
    assert len(result) == 2


@pytest.mark.parametrize("metric",
                         [('ma'),
                          (999),
                          ('xxx')])
def test_auto_naive_metric_value_error(metric):
    train = np.arange(5)
    with pytest.raises(ValueError):
        ms.auto_naive(train, metric=metric)


@pytest.mark.parametrize("method",
                         [('ma'),
                          (999),
                          ('xxx')])
def test_auto_naive_method_value_error(method):
    train = np.arange(5)
    with pytest.raises(ValueError):
        ms.auto_naive(train, method=method)


@pytest.mark.parametrize("window_size",
                         [('ma'),
                          (-999),
                          ('xxx')])
def test_auto_naive_window_size_value_error(window_size):
    train = np.arange(5)
    with pytest.raises(ValueError):
        ms.auto_naive(train, window_size=window_size)


@pytest.mark.parametrize("min_train_size",
                         [('ma'),
                          (-999),
                          ('xxx')])
def test_auto_naive_mintrain_size_value_error(min_train_size):
    train = np.arange(5)
    with pytest.raises(ValueError):
        ms.auto_naive(train, min_train_size=min_train_size)


@pytest.mark.parametrize("train_size, min_train_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_cross_validation_folds_return_length(train_size, min_train_size,
                                              horizon, step, expected):
    y_train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(y_train, min_train_size, horizon, step)
    result = ms.cross_validation_folds(b.Naive1(), cv)

    assert len(result) == expected


@pytest.mark.parametrize("train_size, min_train_size, horizon, step, expected",
                         [(10, 3, 1, 1, 7),
                          (10, 3, 2, 1, 6),
                          (10, 3, 1, 2, 4),
                          (10, 3, 2, 2, 3)])
def test_cross_validation_score_return_length(train_size, min_train_size,
                                              horizon, step, expected):
    y_train = np.arange(train_size)
    cv = ms.rolling_forecast_origin(y_train, min_train_size, horizon, step)
    metric = metrics.mean_error
    result = ms.cross_validation_score(b.Naive1(), cv, metric=metric, n_jobs=1)

    assert len(result) == expected





'''
Unit test for forecast error functions (point and coverage)
in the metrics module
'''

import pytest
import numpy as np

from forecast_tools import metrics as m


@pytest.mark.parametrize("y_true, y_pred, metrics, expected",
                         [([1], [1], 'all', 6),
                          ([1], [1], ['mae'], 1),
                          ([1], [1], ['mae', 'me'], 2),
                          ([1], [1], ['mae', 'me', 'smape'], 3),
                          ([1], [1], ['mae', 'me', 'smape', 'mse',
                                      'rmse', 'mape'], 6)])
def test_forecast_error_return_length(y_true, y_pred, metrics, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    funcs_dict = m.forecast_errors(y_true, y_pred, metrics)
    assert len(funcs_dict) == expected


@pytest.mark.parametrize("y_true, y_pred, metrics, expected",
                         [([1], [1], 'all', ['me', 'mae', 'mse', 'rmse',
                                             'mape', 'smape']),
                          ([1], [1], ['mae'], ['mae']),
                          ([1], [1], ['mae', 'me'], ['mae', 'me']),
                          ([1], [1], ['mae', 'me', 'smape'], ['mae', 'me',
                                                              'smape'])])
def test_forecast_error_return_funcs(y_true, y_pred, metrics, expected):
    '''
    test the correct error functions are returned
    '''
    funcs_dict = m.forecast_errors(y_true, y_pred, metrics)
    assert list(funcs_dict.keys()) == expected


@pytest.mark.parametrize("y_pred, y_true, expected",
                         [([1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], 0.0),
                          ([1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12], 6.0),
                          ([103, 130, 132, 124, 124, 108],
                           [129, 111, 122, 129, 110, 141], 17.833333),
                          ([103, 130, 132, 124, 124, 108, 160, 160],
                           [129, 111, 122, 129, 110, 141, 142, 143], 17.75)])
def test_mean_absolute_error(y_true, y_pred, expected):
    '''
    test mean absolute error calculation
    '''
    error = m.mean_absolute_error(y_true, y_pred)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_pred, y_true, expected",
                         [([1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], 0.0),
                          ([1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12], 6.0),
                          ([103, 130, 132, 124, 124, 108],
                           [129, 111, 122, 129, 110, 141], 3.5),
                          ([103, 130, 132, 124, 124, 108, 160, 160],
                           [129, 111, 122, 129, 110, 141, 142, 143], -1.75)])
def test_mean_error(y_true, y_pred, expected):
    '''
    test mean error calculation
    '''
    error = m.mean_error(y_true, y_pred)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_pred, y_true, expected",
                         [([1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], 0.0),
                          ([1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12],
                           65.3210678210678),
                          ([103, 130, 132, 124, 124, 108],
                           [129, 111, 122, 129, 110, 141],
                           14.2460623711587),
                          ([103, 130, 132, 124, 124, 108, 160, 160],
                           [129, 111, 122, 129, 110, 141, 142, 143],
                           13.7550678066365)])
def test_mean_absolute_percentage_error(y_true, y_pred, expected):
    '''
    test mean error calculation
    '''
    error = m.mean_absolute_percentage_error(y_true, y_pred)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_pred, y_true, expected",
                         [([1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], 0.0),
                          ([1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12],
                           36.0),
                          ([103, 130, 132, 124, 124, 108],
                           [129, 111, 122, 129, 110, 141],
                           407.833333333333),
                          ([103, 130, 132, 124, 124, 108, 160, 160],
                           [129, 111, 122, 129, 110, 141, 142, 143],
                           382.50)])
def test_mean_squared_error(y_true, y_pred, expected):
    '''
    test mean squared error calculation
    '''
    error = m.mean_squared_error(y_true, y_pred)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_pred, y_true, expected",
                         [([1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], 0.0),
                          ([1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12],
                           6.0),
                          ([103, 130, 132, 124, 124, 108],
                           [129, 111, 122, 129, 110, 141],
                           20.1948838405506),
                          ([103, 130, 132, 124, 124, 108, 160, 160],
                           [129, 111, 122, 129, 110, 141, 142, 143],
                           19.5576072156079)])
def test_root_mean_squared_error(y_true, y_pred, expected):
    '''
    test root mean squared error calculation
    '''
    error = m.root_mean_squared_error(y_true, y_pred)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_pred, y_true, expected",
                         [([1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], 0.0),
                          ([1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12],
                           99.5634920634921),
                          ([103, 130, 132, 124, 124, 108],
                           [129, 111, 122, 129, 110, 141],
                           14.7466414897349),
                          ([103, 130, 132, 124, 124, 108, 160, 160],
                           [129, 111, 122, 129, 110, 141, 142, 143],
                           13.9526876064932)])
def test_symmetric_mape(y_true, y_pred, expected):
    '''
    test symmetric mean absolute percentage error calculation
    '''
    error = m.symmetric_mean_absolute_percentage_error(y_true, y_pred)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_true, y_intervals, expected",
                         [([10, 20, 30, 40, 50],
                           [[5, 15, 25, 35, 45],
                            [15, 25, 35, 45, 55]],
                           1.0),
                          ([20, 20, 30, 40, 50],
                           [[5, 15, 25, 35, 45],
                            [15, 25, 35, 45, 55]],
                           0.8),
                          ([20, 30, 30, 40, 50],
                           [[5, 15, 25, 35, 45],
                            [15, 25, 35, 45, 55]],
                           0.6),
                          ([20, 20, 30, 40, 30],
                           [[5, 15, 25, 35, 45],
                            [15, 25, 35, 45, 55]],
                           0.6),
                          ([100, 100, 100, 100, 100],
                           [[5, 15, 25, 35, 45],
                            [15, 25, 35, 45, 55]],
                           0.0)])
def test_coverage(y_true, y_intervals, expected):
    '''
    test prediction interval coverage
    '''
    y_intervals = np.array(y_intervals).T
    error = m.coverage(y_true, y_intervals)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_train, y_pred, y_true, expected",
                         [(np.arange(10), [1, 2, 3, 4, 5, 6],
                          [1, 2, 3, 4, 5, 6], 0.0),
                          (np.arange(1, 21), np.arange(21, 26),
                           np.full(5, 10), 13)])
def test_mase_naive(y_train, y_true, y_pred, expected):
    '''
    test mean absolute scaled error calculation using naive as scaler.

    test calcs produced using libre office calc.
    '''
    error = m.mean_absolute_scaled_error(y_true, y_pred, y_train, period=None)
    assert pytest.approx(expected) == error


@pytest.mark.parametrize("y_train, y_pred, y_true, expected",
                         [(np.arange(1, 21), [1, 2, 3, 4, 5, 6],
                          [1, 2, 3, 4, 5, 6], 0.0),

                          (np.arange(1, 21), np.arange(21, 26),
                           np.full(5, 10), 1.85714286)])
def test_mase_snaive(y_train, y_true, y_pred, expected):
    '''
    test mean absolute scaled error calculation using SNaive as scaler

    test calcs produced using libre office calc.
    '''
    error = m.mean_absolute_scaled_error(y_true, y_pred, y_train, period=7)
    assert pytest.approx(expected) == error



import setuptools
from setuptools.command.egg_info import egg_info


class egg_info_ex(egg_info):
    """Includes license file into `.egg-info` folder."""

    def run(self):
        # don't duplicate license into `.egg-info` when building a distribution
        if not self.distribution.have_run.get('install', True):
            # `install` command is in progress, copy license
            self.mkpath(self.egg_info)
            self.copy_file('LICENSE', self.egg_info)

        egg_info.run(self)

# Read in the requirements.txt file
with open("requirements.txt") as f:
    requirements = []
    for library in f.read().splitlines():
        requirements.append(library)


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="forecast-tools",
    version="0.1.6",
    author="Thomas Monks",
    author_email="forecast_tools@gmail.com",
    license="The MIT License (MIT)",
    license_files=('LICENSE', ),
    description="Tools to support the forecasting process in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TomMonks/forecast-tools",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={"": ["forecast_tools/data/*.csv"]},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.9',
    install_requires=requirements,
)



'''
Tools for simple baseline/benchmark forecasts

These methods might serve as the forecast themselves, but are more likely
to be used as a baseline to evaluate if more complex models offer a sufficient
increase in accuracy to justify their use.

Naive1:
    Carry last value forward across forecast horizon (random walk)

SNaive:
    Carry forward value from last seasonal period

Average: np.sqrt(((h - 1) / self._period).astype(np.int)+1)
    Carry forward average of observations

Drift:
    Carry forward last time period, but allow for upwards/downwards drift.

EnsembleNaive:
    An unweighted average of all of the Naive forecasting methods.
'''

import numpy as np
import pandas as pd
from scipy.stats import norm, t
from abc import ABC, abstractmethod

# Boolean, unsigned integer, signed integer, float, complex.
_NUMERIC_KINDS = set('buifc')


def is_numeric(array):
    """Determine whether the argument has a numeric datatype, when
    converted to a NumPy array.

    Booleans, unsigned integers, signed integers, floats and complex
    numbers are the kinds of numeric datatype.

    source:
    https://codereview.stackexchange.com/questions/128032/check-if-a-numpy-array-contains-numerical-data

    Parameters
    ----------
    array : array-like
        The array to check.

    Returns
    -------
    is_numeric : `bool`
        True if the array has a numeric datatype, False if not.

    """
    return np.asarray(array).dtype.kind in _NUMERIC_KINDS


class Forecast(ABC):
    '''
    Abstract base class for all baseline forecast
    methods
    '''

    def __init__(self):
        self._fitted = None
        self._t = None

    def _get_fitted(self):
        return self._fitted['pred']

    def _get_resid(self):
        return self._fitted['resid']

    @abstractmethod
    def fit(self, train):
        pass

    def fit_predict(self, train, horizon, return_predict_int=False,
                    alpha=None):
        '''
        Convenience method.  Fit model and predict with one call.

        Parameters:
        ---------

        train: array-like,
            vector, series, or dataframe of the time series used for training.
            Values should be floats and not contain any np.nan or np.inf

        horizon: int,
            forecast horizon.

        return_predict_int: bool, optional (default=False)
            If True function will return a Tuple
            0: point forecasts (mean)
            1: matrix of intervals.

        alpha: None, or list of floats, optional (default=None)
            List of floats between 0 and 1. If return_predict_int == True this
            specifies the 100(1-alpha) prediction intervals to return.

        Returns:
        ------
        np.array, vector of predictions. length=horizon

        '''
        self.fit(train)
        return self.predict(horizon, return_predict_int=return_predict_int,
                            alpha=alpha)

    def validate_training_data(self, train, min_length=1):
        '''
        Checks the validity of training data for forecasting
        and raises exceptions if required.

        1. check is instance of pd.Series, pd.DataFrame or np.ndarray
        2. check len is > min_length

        Parameters:
        ---------
        min_length: int optional (default=0)
            minimum length of the time series.

        '''

        if not isinstance(train, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                'Training data must be pd.Series, pd.DataFrame or np.ndarray')
        elif len(train) < min_length:
            raise ValueError('Training data is empty')
        elif not is_numeric(train):
            raise TypeError('Training data must be numeric')

        elif np.isnan(np.asarray(train)).any():
            raise TypeError(
                'Training data contains at least one NaN. '
                + 'Data myst all be floats')
        elif np.isinf(np.asarray(train)).any():
            raise TypeError(
                'Training data contains at least one Infinite '
                + 'value (np.Inf). Data myst all be floats')

    @abstractmethod
    def predict(self, horizon, return_predict_int=False, alpha=None):
        pass

    def _prediction_interval(self, horizon, alpha=None):
        '''
        Prediction intervals for naive forecast 1 (NF1)

        lower = pred - z * std_h
        upper = pred + z * std_h

        where

        std_h = resid_std * sqrt(h)

        resid_std = standard deviation of in-sample residuals

        h = horizon

        See and credit: https://otexts.com/fpp2/prediction-intervals.html

        Pre-requisit: Must have called fit()

        Parameters:
        ---------
        horizon - int,
                    forecast horizon

        levels - list,
            list of floats representing prediction limits
            e.g. [0.80, 0.90, 0.95] will calculate three sets ofprediction
            intervals giving limits for which will include the actual future
            value with probability 80, 90 and 95 percent,
            respectively (default = [0.8, 0.95]).

        Returns:
        --------
        list
            np.array matricies that contain the lower and upper prediction
            limits for each prediction interval specified.

        '''

        if alpha is None:
            alpha = [0.20, 0.05]

        zs = [self.interval_multiplier(1-a, self._t - 1) for a in alpha]

        pis = []

        std_h = self._std_h(horizon)

        for z in zs:
            hw = z * std_h
            pis.append(np.array([self.predict(horizon) - hw,
                                 self.predict(horizon) + hw]).T)

        return pis

    def interval_multiplier(self, level, dof):
        '''
        inverse of normal distribution
        can be overridden if needed.
        '''
        x = norm.ppf((1 - level) / 2)
        return np.abs(x)

    @abstractmethod
    def _std_h(self, horizon):
        '''
        Calculate the standard error of the residuals over
        a forecast horizon.  This is method specific.
        '''
        pass

    # breaks PEP8 to align with statsmodels naming
    fittedvalues = property(_get_fitted)
    resid = property(_get_resid)


class Naive1(Forecast):
    '''
    Naive forecast 1 or NF1: Carry the last value foreward across a
    forecast horizon

    For details and theory see [1]

    Attributes
    ----------
    fittedvalues: pd.Series
        In-sample predictions of training data
    resid: pd.Series
        In-sample residuals

    Methods
    -------
    fit(train)
        fit the model to training data
    predict(horizon, return_predict_int=False, alpha=None)
        Predict h-steps ahead
    fit_predict(train, horizons, return_predict_int=False, alpha=None)
        convenience method.  combine fit() and predict()

    See Also
    --------
    forecast_tools.baseline.SNaive
    forecast_tools.baseline.Drift
    forecast_tools.baseline.Average
    forecast_tools.baseline.EnsembleNaive

    References:
    ----------
    [1]. https://otexts.com/fpp2/simple-methods.html

    Examples:
    --------

    Basic fitting and prediction

    >>> y_train = np.arange(10)
    >>> model = Naive1()
    >>> model.fit(y_train)
    >>> model.predict(horizon=7)
    array([9., 9., 9., 9., 9., 9., 9.]

    fit_predict() convenience method

    >>> y_train = np.arange(10)
    >>> model = Naive1()
    >>> model.fit_predict(y_train, horizon=7)
    array([9., 9., 9., 9., 9., 9., 9.]

    80 and 95% prediction intervals

    >>> y_train = np.arange(10)
    >>> model = Naive1()
    >>> model.fit(y_train)
    >>> y_pred, y_intervals = model.predict(horizon=2,
                                            return_pred_interval=True,
                                            alpha=[0.1, 0.05])
    >>> y_pred
    array([9., 9.]
    >>> y_intervals[0]
    array([[ 7.71844843, 10.28155157],
           [ 7.1876124 , 10.8123876 ]])
    >>> y_intervals[1]
    array([[ 7.35514637, 10.64485363],
           [ 6.67382569, 11.32617431]])


    Fitted values (one step in-sample predictions)
    .fittedvalue returns a pandas.Series called pred

    >>> y_train = np.arange(5)
    >>> model = Naive1()
    >>> model.fit(y_train)
    >>> model.fittedvalues
    0    NaN
    1    0.0
    2    1.0
    3    2.0
    4    3.0
    Name: pred, dtype: float64

    '''
    def __init__(self):
        '''
        Constructor method

        Parameters:
        -------
        level - list,
            confidence levels for prediction intervals (e.g. [90, 95])
        '''
        self._fitted = None

    def __repr__(self):
        '''
        String representation of object
        '''
        return f'Naive1()'

    def __str__(self):
        '''
        Print/str representation of object
        '''
        return f'Naive1()'

    def fit(self, train):
        '''
        Train the naive model

        Parameters:
        --------
        train - array-like,
            vector, series, or dataframe of the time series used for training.
            Values should be floats and not contain any np.nan or np.inf
        '''

        self.validate_training_data(train)

        _train = np.asarray(train)
        self._pred = _train[-1]
        self._fitted = pd.DataFrame(_train)

        if isinstance(train, (pd.DataFrame, pd.Series)):
            self._fitted.index = train.index

        self._t = len(_train)
        self._fitted.columns = ['actual']
        self._fitted['pred'] = self._fitted['actual'].shift(periods=1)
        self._fitted['resid'] = self._fitted['actual'] - self._fitted['pred']
        self._resid_std = np.sqrt(np.nanmean(np.square(self._fitted['resid'])))

    def predict(self, horizon, return_predict_int=False, alpha=None):
        '''
        Forecast and optionally produce 100(1-alpha) prediction intervals.

        Prediction intervals for naive forecast 1 (NF1)

        lower = pred - z * std_h
        upper = pred + z * std_h

        where

        std_h = resid_std * sqrt(h)

        resid_std = standard deviation of in-sample residuals

        h = horizon

        See and credit: https://otexts.com/fpp2/prediction-intervals.html

        Pre-requisit: Must have called fit()

        Parameters:
        --------
        horizon - int,
                    forecast horizon.

        return_predict_int: bool, optional
                    if True calculate 100(1-alpha) prediction
                intervals for the forecast. (default=False)

        alpha: list of floats, optional (default=None)
            controls set of prediction intervals returned and the width of
            each.

            Intervals are 100(1-alpha) in width. e.g. [0.2, 0.1]
            would return the 80% and 90% prediction intervals of the forecast
            distribution.  default=None.  When return_predict_int = True the
            default behaviour is to return 80 and 90% intervals.

        Returns:
        -------

        if return_predict_int = False

        np.array, vector of predictions. length=horizon

        if return_predict_int = True then returns a tuple.

        0. np.array, vector of predictions. length=horizon
        1. list of numpy.array[lower_pi, upper_pi].
            One for each prediction interval.

        '''
        if self._fitted is None:
            raise UnboundLocalError('Must call fit() prior to predict()')

        if alpha is None:
            alpha = [0.2, 0.1]

        preds = np.full(shape=horizon, fill_value=self._pred, dtype=float)

        if return_predict_int:
            return preds, self._prediction_interval(horizon, alpha)
        else:
            return preds

    def _std_h(self, horizon):
        '''
        Calculate the sample standard deviation.
        '''
        indexes = np.sqrt(np.arange(1, horizon+1))

        std = np.full(shape=horizon,
                      fill_value=self._resid_std,
                      dtype=np.float)

        std_h = std * indexes
        return std_h


class SNaive(Forecast):
    '''
    Seasonal Naive Forecast SNF

    Each forecast to be equal to the last observed value from the
    same season of the year (e.g., the same month of the previous year).

    SNF is useful for highly seasonal data. See [1]

    Attributes
    ----------
    fittedvalues: pd.Series
        In-sample predictions of training data
    resid: pd.Series
        In-sample residuals

    Methods
    -------
    fit(train)
        fit the model to training data
    predict(horizon, return_predict_int=False, alpha=None)
        Predict h-steps ahead
    fit_predict(train, horizons, return_predict_int=False, alpha=None)
        convenience method.  combine fit() and predict()

    See Also
    --------
    forecast_tools.baseline.Naive1
    forecast_tools.baseline.Drift
    forecast_tools.baseline.Average
    forecast_tools.baseline.EnsembleNaive

    References:
    -----------
    [1]. https://otexts.com/fpp2/simple-methods.html

    '''

    def __init__(self, period):
        '''
        Parameters:
        --------
        period - int, the seasonal period of the daya
                 e.g. weekly = 7, monthly = 12, daily = 24
        '''
        self._period = period
        self._fitted = None

    def __repr__(self):
        '''
        String representation of object
        '''
        return f'SNaive1(period={self._period})'

    def __str__(self):
        '''
        Print/str representation of object
        '''
        return f'SNaive1(period={self._period})'

    def fit(self, train):
        '''
        Seasonal naive forecast - train the model

        Parameters:
        --------
        train: array-like.
            vector, pd.DataFrame or pd.Series containing the time series used
            for training. Values should be floats and not contain any np.nan
            or np.inf
        '''

        self.validate_training_data(train, min_length=self._period)

        # could refactor this to be more like Naive1's simpler implementation.
        if isinstance(train, (pd.Series)):
            self._f = np.asarray(train)[-self._period:]
            _train = train.to_numpy()
            self._fitted = pd.DataFrame(_train, index=train.index)
        elif isinstance(train, (pd.DataFrame)):
            self._f = train.to_numpy().T[0][-self._period:]
            _train = train.copy()[train.columns[0]].to_numpy()
            self._fitted = pd.DataFrame(_train, index=train.index)
        else:
            self._f = train[-self._period:]
            _train = train.copy()
            self._fitted = pd.DataFrame(_train)

        self._t = len(_train)
        self._fitted.columns = ['actual']
        self._fitted['pred'] = self._fitted['actual'].shift(self._period)
        self._fitted['resid'] = self._fitted['actual'] - self._fitted['pred']

        self._resid_std = np.sqrt(np.nanmean(np.square(self._fitted['resid'])))

    def predict(self, horizon, return_predict_int=False, alpha=None):
        '''
        Predict time series over a horizon

        Parameters:
        --------
        horizon - int,
            forecast horizon.

        return_predict_int: bool, optional
                    if True calculate 100(1-alpha) prediction
                intervals for the forecast. (default=False)

        alpha: list of floats, optional (default=None)
            controls set of prediction intervals returned and the width of
            each.

            Intervals are 100(1-alpha) in width. e.g. [0.2, 0.1]
            would return the 80% and 90% prediction intervals of the forecast
            distribution.  default=None.  When return_predict_int = True the
            default behaviour is to return 80 and 90% intervals.

        Returns:
        --------

        if return_predict_int = False

        np.array, vector of predictions. length=horizon

        if return_predict_int = True then returns a tuple.

        0. np.array, vector of predictions. length=horizon
        1. list of numpy.array[lower_pi, upper_pi].
            One for each prediction interval.
        '''

        if self._fitted is None:
            raise UnboundLocalError('Must call fit() prior to predict()')

        if alpha is None:
            alpha = [0.2, 0.1]

        preds = np.array([], dtype=float)

        for _ in range(0, int(horizon/self._period)):
            preds = np.concatenate([preds, self._f.copy()], axis=0)

        preds = np.concatenate([preds,
                               self._f.copy()[:horizon % self._period]],
                               axis=0)

        if return_predict_int:
            return preds, self._prediction_interval(horizon, alpha)
        else:
            return preds

    def _std_h(self, horizon):
        h = np.arange(1, horizon+1)
        # need to query if should be +1 or not.
        return self._resid_std * \
            np.sqrt(((h - 1) / self._period).astype(np.int)+1)


class Average(Forecast):
    '''
    Average forecast.  Forecast is set to the average
    of the historical data.

    See for discussion of the average as a forecat measure [1]

    Attributes
    ----------
    fittedvalues: pd.Series
        In-sample predictions of training data
    resid: pd.Series
        In-sample residuals

    Methods
    -------
    fit(train)
        fit the model to training data
    predict(horizon, return_predict_int=False, alpha=None)
        Predict h-steps ahead
    fit_predict(train, horizons, return_predict_int=False, alpha=None)
        convenience method.  combine fit() and predict()

    See Also
    --------
    forecast_tools.baseline.Naive1
    forecast_tools.baseline.SNaive
    forecast_tools.baseline.Drift
    forecast_tools.baseline.EnsembleNaive

    References:
    -----------
    [1.] Makridakis, Wheelwright and Hyndman. Forecasting (1998)
    '''

    def __init__(self):
        self._pred = None
        self._fitted = None

    def __repr__(self):
        '''
        String representation of object
        '''
        return f'Average()'

    def __str__(self):
        '''
        Print/str representation of object
        '''
        return f'Average()'

    def _get_fitted(self):
        return self._fitted['pred']

    def _get_resid(self):
        return self._fitted['resid']

    def fit(self, train):
        '''
        Train the model

        Parameters:
        --------
        train:  arraylike
                vector, pd.series, pd.DataFrame,
                Time series used for training.  Values should be floats
                and not contain any np.nan or np.inf
        '''

        self.validate_training_data(train)

        if isinstance(train, (pd.DataFrame)):
            _train = train.copy()[train.columns[0]].to_numpy()
            self._fitted = pd.DataFrame(_train, index=train.index)
        elif isinstance(train, (pd.Series)):
            _train = train.to_numpy()
            self._fitted = pd.DataFrame(_train, index=train.index)
        else:
            _train = train.copy()
            self._fitted = pd.DataFrame(train)

        self._fitted.columns = ['actual']

        self._t = len(_train)
        self._pred = _train.mean()
        # ddof set to get sample mean
        self._resid_std = (_train - self._pred).std(ddof=1)
        self._fitted['pred'] = self._pred
        self._fitted['resid'] = self._fitted['actual'] - self._fitted['pred']

    def predict(self, horizon, return_predict_int=False, alpha=None):
        '''
        Predict time series over a horizon

        Parameters:
        --------
        horizon - int, forecast horizon.

        return_predict_int: bool, optional
                    if True calculate 100(1-alpha) prediction
                intervals for the forecast. (default=False)

        alpha: list of floats, optional (default=None)
            controls set of prediction intervals returned and the width of
            each.

            Intervals are 100(1-alpha) in width. e.g. [0.2, 0.1]
            would return the 80% and 90% prediction intervals of the forecast
            distribution.  default=None.  When return_predict_int = True the
            default behaviour is to return 80 and 90% intervals.


        Returns:
        --------

        if return_predict_int = False

        np.array, vector of predictions. length=horizon

        if return_predict_int = True then returns a tuple.

        0. np.array, vector of predictions. length=horizon
        1. list of numpy.array[lower_pi, upper_pi].
            One for each prediction interval.
        '''

        if self._fitted is None:
            raise UnboundLocalError('Must call fit() prior to predict()')

        if alpha is None:
            alpha = [0.2, 0.1]

        preds = np.full(shape=horizon, fill_value=self._pred, dtype=float)

        if return_predict_int:
            return preds, self._prediction_interval(horizon, alpha)
        else:
            return preds

    def interval_multiplier(self, level, dof):
        '''
        inverse of student t distribution
        '''
        x = t.ppf((1 - level) / 2, dof)
        return np.abs(x)

    def _std_h(self, horizon):
        std = self._resid_std * np.sqrt(1 + (1/self._t))
        return np.full(shape=horizon, fill_value=std, dtype=np.float)


class Drift(Forecast):
    '''
    Naive1 forecast with drift

    Carry the last value foreward across a forecast horizon but
    allow for upwards of downwards drift defined in [1]

    Drift = average change in the historical data.

    Note.  The current implementation has a standard error of the forecast
    that is the same as for the naive1 se.  This could be adjusted for drift.
    The following link suggests this is minor and benchmark with R is v.similar
    [2]

    Attributes
    ----------
    fittedvalues: pd.Series
        In-sample predictions of training data
    resid: pd.Series
        In-sample residuals

    Methods
    -------
    fit(train)
        fit the model to training data
    predict(horizon, return_predict_int=False, alpha=None)
        Predict h-steps ahead
    fit_predict(train, horizons, return_predict_int=False, alpha=None)
        convenience method.  combine fit() and predict()

    See Also
    --------
    forecast_tools.baseline.Naive1
    forecast_tools.baseline.SNaive
    forecast_tools.baseline.Average
    forecast_tools.baseline.EnsembleNaive

    References:
    -----------
    [1]. https://otexts.com/fpp2/simple-methods.html
    [2]. https://www.coursehero.com/file/p12k3ln/For-the-random-walk-with-drift-model-the-1-step-ahead-forecast-standard-error/

    '''

    def __init__(self):
        self._fitted = None

    def __repr__(self):
        '''
        String representation of object
        '''
        return f'Drift()'

    def __str__(self):
        '''
        Print/str representation of object
        '''
        return f'Drift()'

    def _get_fitted_gradient(self):
        return self._fitted['gradient_fit']

    def fit(self, train):
        '''
        Train the naive with drift model

        Parameters:
        --------
        train:  arraylike
                vector, pd.series, pd.DataFrame,
                Time series used for training.  Values should be floats
                and not contain any np.nan or np.inf

        '''

        self.validate_training_data(train)

        # if dataframe convert to series for compatability with
        # proc (for convenience of passing the dataframe rather than a series)
        if isinstance(train, (pd.DataFrame)):
            _train = train.copy()[train.columns[0]].to_numpy()
            self._fitted = pd.DataFrame(_train, index=train.index)
        elif isinstance(train, (pd.Series)):
            _train = train.to_numpy()
            self._fitted = pd.DataFrame(_train, index=train.index)
        else:
            _train = train.copy()
            self._fitted = pd.DataFrame(train)

        self._last_value = _train[-1:][0]
        self._t = _train.shape[0]
        self._gradient = ((self._last_value - _train[0]) / (self._t - 1))
        self._fitted.columns = ['actual']

        # could show fitted as line from first to last point.
        # unclear if should use this or naive1 method.
        self._fitted['gradient_fit'] = _train[0] \
            + np.arange(1, self._t+1, dtype=float) * self._gradient

        # 1 step carry forward naive1 fitted values.
        self._fitted['pred'] = self._fitted['actual'].shift(periods=1)
        self._fitted['resid'] = self._fitted['actual'] - self._fitted['pred']

        # standard error is not adjusted for drift - how much of an issue?
        self._resid_std = np.sqrt(np.nanmean(np.square(self._fitted['resid'])))

    def predict(self, horizon, return_predict_int=False, alpha=None):
        '''
        Parameters:
        --------
        horizon - int, forecast horizon.

        return_predict_int: bool, optional
                    if True calculate 100(1-alpha) prediction
                intervals for the forecast. (default=False)

        alpha: list of floats, optional (default=None)
            controls set of prediction intervals returned and the width of
            each.

            Intervals are 100(1-alpha) in width. e.g. [0.2, 0.1]
            would return the 80% and 90% prediction intervals of the forecast
            distribution.  default=None.  When return_predict_int = True the
            default behaviour is to return 80 and 90% intervals.


        Returns:
        --------

        if return_predict_int = False

        np.array, vector of predictions. length=horizon

        if return_predict_int = True then returns a tuple.

        0. np.array, vector of predictions. length=horizon
        1. list of numpy.array[lower_pi, upper_pi].
            One for each prediction interval.
        '''

        if self._fitted is None:
            raise UnboundLocalError('Must call fit() prior to predict()')

        if alpha is None:
            alpha = [0.2, 0.1]

        preds = np.arange(1, horizon+1, dtype=float) * self._gradient
        preds += self._last_value

        if return_predict_int:
            return preds, self._prediction_interval(horizon, alpha)
        else:
            return preds

    def _std_h(self, horizon):

        h = np.arange(1, horizon+1)
        return self._resid_std * np.sqrt(h * (1 + (h / self._t)))

    fitted_gradient = property(_get_fitted_gradient)


class EnsembleNaive(Forecast):
    '''
    An ensemble of all naive forecast methods.

    Attributes
    ----------
    fittedvalues: pd.Series
        In-sample predictions of training data
    resid: pd.Series
        In-sample residuals

    Methods
    -------
    fit(train)
        fit the model to training data
    predict(horizon, return_predict_int=False, alpha=None)
        Predict h-steps ahead
    fit_predict(train, horizons, return_predict_int=False, alpha=None)
        convenience method.  combine fit() and predict()

    See Also
    --------
    forecast_tools.baseline.Naive1
    forecast_tools.baseline.SNaive
    forecast_tools.baseline.Average
    forecast_tools.baseline.Drift
    '''

    def __init__(self, seasonal_period):
        self._estimators = {'NF1': Naive1(),
                            'SNaive': SNaive(period=seasonal_period),
                            'Average': Average(),
                            'Drift': Drift()
                            }

    def __repr__(self):
        '''
        String representation of object
        '''
        p = self._estimators['SNaive']._period
        return f'EnsembleNaive(seasonal_period={p})'

    def __str__(self):
        '''
        Print/str representation of object
        '''
        p = self._estimators['SNaive']._period
        return f'EnsembleNaive(seasonal_period={p})'

    def fit(self, train):
        '''
        Parameters:
        --------

        train:  arraylike
                vector, pd.series, pd.DataFrame,
                Time series used for training.  Values should be floats
                and not contain any np.nan or np.inf
        '''

        for _, estimator in self._estimators.items():
            estimator.fit(train)

    def predict(self, horizon, return_predict_int=False, alpha=None):
        preds = []
        for _, estimator in self._estimators.items():
            model_preds = estimator.predict(horizon)
            preds.append(model_preds)

        return np.array(preds).mean(axis=0)

    def _std_h(self, horizon):
        '''
        Calculate the standard error of the residuals over
        a forecast horizon.  This is method specific.
        '''
        pass


def baseline_estimators(seasonal_period):
    '''
    Generate a collection of baseline forecast objects

    Parameters:
    --------
    seasonal_period - int,
        order of seasonal periods in the data (e.g daily = 7)

    average_lookback - int,
        number of lagged periods that average baseline includes

    Returns:
    --------
    dict,
        forecast objects
    '''

    estimators = {'NF1': Naive1(),
                  'SNaive': SNaive(period=seasonal_period),
                  'Average': Average(),
                  'Drift': Drift(),
                  'Ensemble': EnsembleNaive(seasonal_period=seasonal_period)
                  }

    return estimators


def boot_prediction_intervals(preds, resid, horizon, levels=None, boots=1000):
    '''
    Constructs bootstrap prediction intervals for forecasting.

    This procedure makes no assumptions about the distribution of errors
    e.g if they are normally distributed, but does assumes that no significant
    autocorrelation exists in residuals.

    Parameters:
    -----------

    preds: array-like,
        predictions over forecast horizon

    resid: array-like,
        in-sample prediction residuals

    horizon: int,
        forecast horizon (e.g. 12 months or 7 days)

    levels: list of floats,
        prediction interval precisions (default=[0.80, 0.95])

    boots: int,
        number of bootstrap datasets to construct (default = 1000)

    Returns:
    ---------

    list of numpy arrays.  Each numpy array contains two columns of the upper
    and lower prediction limits across the forecast horizon.
    '''

    if levels is None:
        levels = [0.80, 0.95]

    resid = _drop_na_from_series(resid)

    sample = np.random.choice(resid, size=(boots, horizon))
    sample = np.cumsum(sample, axis=1)

    data = preds + sample

    pis = []
    for level in levels:

        alpha = (1 - level) / 2
        q_upper = level + alpha
        q_lower = (1 - level) - alpha

        upper = np.percentile(data, q_upper*100, interpolation='higher',
                              axis=0)
        lower = np.percentile(data, q_lower*100, interpolation='higher',
                              axis=0)

        pis.append(np.array([lower, upper]).T)

    return pis


def _drop_na_from_series(data):
    '''
    Drops all NaN from numpy array or pandas series.

    Parameters:
    -------
    data, array-like,
    np.ndarray or pd.Series.

    Returns:
    -------
    np.ndarray removing NaN.

    '''
    if isinstance(data, pd.Series):
        return data.dropna().to_numpy()
    else:
        return data[~np.isnan(data)]



'''
Tests for Naive benchmark classes

Tests currently cover:

1. Forecast horizons
2. Allowable input types: np.ndarray, pd.DataFrame, pd.Series
3. Failure paths for abnormal input such as np.nan, non numeric,
   empty arrays and np.Inf
4. Predictions
    - naive1 - carries forward last value
    - snaive - carries forward previous h values
    - average - flat forecast of average
    - drift - previous value + gradient
    - ensemble naive - the average of all of the methods
    - Test fit_predict()

5. Prediction intervals
    - horizon
    - sets i.e. 2 sets of intervals (0.8 and 0.95)
    - width
    - bootstrapped prediction intervals
        - length of horizon
        - number of sets of intervals returned.

6. Fitted values
    - expected length
    - count of NaN
'''

import pytest
import pandas as pd
import numpy as np

import forecast_tools.baseline as b


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_naive1_forecast_horizon(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Naive1()
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_naive1_fit_predict(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Naive1()
    # fit_predict for point forecasts only
    preds = model.fit_predict(pd.Series(data), horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_snaive_forecast_horizon(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.SNaive(1)
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_snaive_fit_predict(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.SNaive(1)
    # fit_predict for point forecasts only
    preds = model.fit_predict(pd.Series(data), horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_drift_forecast_horizon(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Drift()
    model.fit(np.array(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_drift_fit_predict(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Drift()
    # fit_predict for point forecasts only
    preds = model.fit_predict(pd.Series(data), horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_average_forecast_horizon(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Average()
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_average_fit_predict(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Average()
    # fit_predict for point forecasts only
    preds = model.fit_predict(pd.Series(data), horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_average_forecast_input_numpy(data, horizon, expected):
    '''
    test the average class accepts numpy array
    '''
    model = b.Average()
    model.fit(np.array(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_average_forecast_input_series(data, horizon, expected):
    '''
    test the average class accepts pandas series
    '''
    model = b.Average()
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_average_forecast_input_dataframe(data, horizon, expected):
    '''
    test the average baseline class accept dataframe
    '''
    model = b.Average()
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_naive1_forecast_input_dataframe(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Naive1()
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_naive1_forecast_input_series(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Naive1()
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_naive1_forecast_input_numpy(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Naive1()
    model.fit(np.array(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_snaive_forecast_input_series(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.SNaive(1)
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_snaive_forecast_input_dataframe(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.SNaive(1)
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_drift_forecast_input_numpy(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Drift()
    model.fit(np.array(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_drift_forecast_input_series(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Drift()
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_drift_forecast_input_dataframe(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Drift()
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_ensemble_forecast_input_dataframe(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.EnsembleNaive(2)
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_ensemble_forecast_input_series(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.EnsembleNaive(2)
    model.fit(pd.Series(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, horizon, expected",
                         [([1, 2, 3, 4, 5], 12, 12),
                          ([1, 2, 3, 4, 5], 24, 24),
                          ([1, 2, 3], 8, 8)
                          ])
def test_ensemble_forecast_input_numpy(data, horizon, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.EnsembleNaive(2)
    model.fit(np.array(data))
    # point forecasts only
    preds = model.predict(horizon)
    assert len(preds) == expected


@pytest.mark.parametrize("data, exception",
                         [(np.array([]), ValueError),
                          (1.0, TypeError),
                          (np.array(['foo', 'bar', 'spam', 'eggs']),
                          TypeError),
                          (np.array([1, 2, 3, 4, 5, 6, np.NAN]), TypeError),
                          (np.array([1, 2, 3, 4, np.Inf, 5, 6]), TypeError)])
def test_ensemble_abnormal_input(data, exception):
    '''
    test naive1 raises correct exceptions on abnormal input
    '''
    model = b.EnsembleNaive(2)
    with pytest.raises(exception):
        model.fit(data)


@pytest.mark.parametrize("data, expected",
                         [([1, 2, 3, 4, 5], 3.0),
                          ([139,  32,  86, 123,  61,  51, 108,
                           137,  33,  25], 79.5),
                          ([1, 2, 3], 2.0)
                          ])
def test_average_forecast_output(data, expected):
    '''
    test the correct number of error metric functions are returned.
    '''
    model = b.Average()
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(1)
    assert preds[0] == expected


@pytest.mark.parametrize("data, expected",
                         [([1, 2, 3, 4, 5], 5.0),
                          ([139,  32,  86, 123,  61,  51,
                           108, 137,  33,  25], 25.0),
                          ([1, 2, 3], 3.0)
                          ])
def test_naive1_forecast_output(data, expected):
    '''
    test naive1 carries forward the last value in the series
    '''
    model = b.Naive1()
    model.fit(pd.DataFrame(data))
    # point forecasts only
    preds = model.predict(1)
    assert preds[0] == expected


@pytest.mark.parametrize("data, period, expected",
                         [(np.resize(np.arange(12), 24), 12, np.arange(12)),
                          (np.resize(np.arange(24), 48), 24, np.arange(24)),
                          (pd.Series(np.resize(np.arange(12), 24)),
                           12, pd.Series(np.arange(12)))
                          ])
def test_snaive_forecast_output(data, period, expected):
    '''
    test naive1 carries forward the last value in the series
    '''
    model = b.SNaive(period)
    model.fit(data)
    # point forecasts only
    preds = model.predict(period)
    assert np.array_equal(preds, expected)


@pytest.mark.parametrize("data, period, expected",
                         [(np.resize(np.arange(12), 24), 12, np.full(12,
                           np.arange(12).mean())),
                          (np.resize(np.arange(24), 48), 24,
                           np.full(24, np.arange(24).mean())),
                          (pd.Series(np.resize(np.arange(12), 24)),
                           12, np.full(12, np.arange(12).mean()))
                          ])
def test_average_forecast_output_longer_horizon(data, period, expected):
    '''
    test naive1 carries forward the last value in the series
    '''
    model = b.Average()
    model.fit(data)
    # point forecasts only
    preds = model.predict(period)
    assert np.array_equal(preds, expected)


@pytest.mark.parametrize("data, exception",
                         [(np.array([]), ValueError),
                          (1.0, TypeError),
                          (np.array(['foo', 'bar', 'spam', 'eggs']),
                          TypeError),
                          (np.array([1, 2, 3, 4, 5, 6, np.NAN]), TypeError),
                          (np.array([1, 2, 3, 4, np.Inf, 5, 6]), TypeError)])
def test_naive1_abnormal_input(data, exception):
    '''
    test naive1 raises correct exceptions on abnormal input
    '''
    model = b.Naive1()
    with pytest.raises(exception):
        model.fit(data)


@pytest.mark.parametrize("data, exception",
                         [(np.array([]), ValueError),
                          (1.0, TypeError),
                          (np.array(['foo', 'bar', 'spam', 'eggs']),
                          TypeError),
                          (np.array([1, 2, 3, 4, 5, 6, np.nan]), TypeError),
                          (np.array([1, 2, 3, 4, np.Inf, 5, 6]), TypeError)
                          ])
def test_snaive_abnormal_input(data, exception):
    '''
    test snaive raises correct exceptions on abnormal input
    '''
    model = b.SNaive(1)
    with pytest.raises(exception):
        model.fit(data)


@pytest.mark.parametrize("data, exception",
                         [(np.array([]), ValueError),
                          (1.0, TypeError),
                          (np.array(['foo', 'bar', 'spam', 'eggs']),
                          TypeError),
                          (np.array([1, 2, 3, 4, 5, 6, np.nan]), TypeError),
                          (np.array([1, 2, 3, 4, np.Inf, 5, 6]), TypeError)])
def test_average_abnormal_input(data, exception):
    '''
    test average raises correct exceptions on abnormal input
    '''
    model = b.Average()
    with pytest.raises(exception):
        model.fit(data)


@pytest.mark.parametrize("data, exception",
                         [(np.array([]), ValueError),
                          (1.0, TypeError),
                          (np.array(['foo', 'bar', 'spam', 'eggs']),
                          TypeError),
                          (np.array([1, 2, 3, 4, 5, 6, np.nan]), TypeError),
                          (np.array([1, 2, 3, 4, np.Inf, 5, 6]), TypeError)])
def test_drift_abnormal_input(data, exception):
    '''
    test drift raises correct exceptions on abnormal input
    '''
    model = b.Drift()
    with pytest.raises(exception):
        model.fit(data)


@pytest.mark.parametrize("data, horizon, alpha, expected",
                         [([1, 2, 3, 4, 5], 12, [0.2, 0.05], 12),
                          ([1, 2, 3, 4, 5], 24, [0.2, 0.10, 0.05], 24),
                          ([1, 2, 3], 8, [0.8], 8)
                          ])
def test_naive1_pi_horizon(data, horizon, alpha, expected):
    '''
    test the correct forecast horizon is returned for prediction interval
    for Naive1
    '''
    model = b.Naive1()
    model.fit(pd.Series(data))
    # point forecasts only
    _, intervals = model.predict(horizon, return_predict_int=True, alpha=alpha)
    assert len(intervals[0]) == expected


@pytest.mark.parametrize("data, horizon, alpha, expected",
                         [([1, 2, 3, 4, 5], 12, [0.2, 0.05], 12),
                          ([1, 2, 3, 4, 5], 24, [0.2, 0.10, 0.05], 24),
                          ([1, 2, 3], 8, [0.8], 8)
                          ])
def test_snaive_pi_horizon(data, horizon, alpha, expected):
    '''
    test the correct forecast horizon is returned for prediction
    interval for SNaive
    '''
    model = b.SNaive(1)
    model.fit(pd.Series(data))
    # point forecasts only
    _, intervals = model.predict(horizon, return_predict_int=True, alpha=alpha)
    assert len(intervals[0]) == expected


@pytest.mark.parametrize("data, horizon, alpha, expected",
                         [([1, 2, 3, 4, 5], 12, [0.2, 0.05], 12),
                          ([1, 2, 3, 4, 5], 24, [0.2, 0.10, 0.05], 24),
                          ([1, 2, 3], 8, [0.8], 8)
                          ])
def test_drift_pi_horizon(data, horizon, alpha, expected):
    '''
    test the correct forecast horizon is returned for prediction
    interval for Drift
    '''
    model = b.Drift()
    model.fit(pd.Series(data))
    # point forecasts only
    _, intervals = model.predict(
        horizon, return_predict_int=True, alpha=alpha)
    assert len(intervals[0]) == expected


@pytest.mark.parametrize("data, horizon, alpha, expected",
                         [([1, 2, 3, 4, 5], 12, [0.2, 0.05], 12),
                          ([1, 2, 3, 4, 5], 24, [0.2, 0.10, 0.05], 24),
                          ([1, 2, 3], 8, [0.8], 8)
                          ])
def test_average_pi_horizon(data, horizon, alpha, expected):
    '''
    test the correct forecast horizon is returned for prediction
    interval for Average
    '''
    model = b.Average()
    model.fit(pd.Series(data))
    # point forecasts only
    _, intervals = model.predict(
        horizon, return_predict_int=True, alpha=alpha)
    assert len(intervals[0]) == expected


@pytest.mark.parametrize("model, data, horizon, alpha, expected",
                         [(b.Naive1(), [1, 2, 3, 4, 5], 12, [0.2, 0.05], 2),
                          (b.Naive1(), [1, 2, 3, 4, 5],
                           24, [0.2, 0.10, 0.05], 3),
                          (b.SNaive(1), [1, 2, 3], 8, [0.8], 1),
                          (b.SNaive(1), [1, 2, 3, 4, 5],
                           24, [0.2, 0.10, 0.05], 3),
                          (b.Naive1(), [1, 2, 3], 8, None, 2),
                          (b.SNaive(1), [1, 2, 3], 8, None, 2),
                          (b.Average(), [1, 2, 3], 8, None, 2),
                          (b.Drift(), [1, 2, 3], 8, None, 2),
                          (b.Drift(), [1, 2, 3], 8, [0.8], 1),
                          (b.Drift(), [1, 2, 3], 8, None, 2),
                          (b.Average(), [1, 2, 3, 4, 5],
                           24, [0.2, 0.10, 0.05], 3)
                          ])
def test_naive_pi_set_number(model, data, horizon, alpha, expected):
    '''
    test the correct number of Prediction intervals are
    returned for prediction interval for all Naive forecasting classes
    '''
    model.fit(pd.Series(data))
    # point forecasts only
    _, intervals = model.predict(
        horizon, return_predict_int=True, alpha=alpha)
    assert len(intervals) == expected


@pytest.mark.parametrize("data, period, expected",
                         [(np.arange(1, 7), 6, np.arange(7, 13)),
                          (pd.Series(np.arange(1, 7)), 6, np.arange(7, 13)),
                          (pd.DataFrame(np.arange(1, 7)), 6, np.arange(7, 13)),
                          (pd.DataFrame(np.arange(1.0, 7.0,
                                        dtype=np.float64)), 6,
                           np.arange(7.0, 13.0, dtype=np.float64))
                          ])
def test_drift_forecast_output_longer_horizon(data, period, expected):
    '''
    test drift forecast predictions
    '''
    model = b.Drift()
    model.fit(data)
    # point forecasts only
    preds = model.predict(period)
    assert np.array_equal(preds, expected)


def test_naive1_prediction_interval_low():
    '''
    test naive 80% lower prediction interval
    '''

    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    low = [29.56885, 24.005, 19.73657, 16.13770, 12.96704]
    # high = [56.43115, 61.99451, 66.26343, 69.86230, 73.03296]

    model = b.Naive1()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.2])
    print(intervals[0].T[0])
    assert pytest.approx(intervals[0].T[0], rel=1e-6, abs=0.1) == low


def test_naive1_prediction_interval_high():
    '''
    test naive 80% upper prediction interval
    '''

    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [29.56885, 24.005, 19.73657, 16.13770, 12.96704]
    high = [56.43115, 61.99451, 66.26343, 69.86230, 73.03296]

    model = b.Naive1()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.2])

    print(intervals[0].T[1])
    assert pytest.approx(intervals[0].T[1], rel=1e-6, abs=0.1) == high


def test_naive1_se():
    '''
    standard error of naive1 is root mean squared.
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [29.56885, 24.005, 19.73657, 16.13770, 12.96704]
    # high = [56.43115, 61.99451, 66.26343, 69.86230, 73.03296]

    model = b.Naive1()
    model.fit(train)

    expected = 10.48038

    assert pytest.approx(model._resid_std) == expected


def test_average_prediction_interval_high():
    '''
    test average 80% upper prediction interval
    '''

    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [40.97369, 40.97369, 40.97369, 40.97369, 40.97369]
    high = [59.34631, 59.34631, 59.34631, 59.34631, 59.34631]

    model = b.Average()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.2])

    print(intervals[0].T[1])
    # assert np.array_equal(intervals[0].T[1], high)
    assert pytest.approx(intervals[0].T[1]) == high


def test_average_prediction_interval_low():
    '''
    test average 80% lower prediction interval
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    low = [40.97369, 40.97369, 40.97369, 40.97369, 40.97369]
    # high = [59.34631, 59.34631, 59.34631, 59.34631, 59.34631]

    model = b.Average()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.2])

    print(intervals[0].T[1])

    assert pytest.approx(intervals[0].T[0]) == low


def test_naive1_prediction_interval_95_high():
    '''
    test naive1 95% upper prediction interval
    '''

    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [22.458831, 13.950400, 7.421651, 1.917662, -2.931450]
    high = [63.54117, 72.04960, 78.57835, 84.08234, 88.93145]

    model = b.Naive1()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.05])

    print(intervals[0].T[1])
    assert pytest.approx(intervals[0].T[1], rel=1e-6, abs=0.1) == high


def test_naive1_prediction_interval_95_low():
    '''
    test naive1 95% lower prediction interval
    '''

    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    low = [22.458831, 13.950400, 7.421651, 1.917662, -2.931450]
    # high = [63.54117, 72.04960, 78.57835, 84.08234, 88.93145]

    model = b.Naive1()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.05])

    print(intervals[0].T[0])
    assert pytest.approx(intervals[0].T[0], rel=1e-6, abs=0.1) == low


def test_snaive_prediction_interval_80_low():
    '''
    test snaive 80% lower prediction interval
    intervals are matched from R forecast package
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    low = [32.00420, 57.00420, 49.00420, 30.00420, 26.62116]
    # high = [57.99580, 82.99580, 74.99580, 55.99580, 63.37884]

    # quarterly data
    model = b.SNaive(period=4)
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.2])

    print(intervals[0].T[0])
    assert pytest.approx(intervals[0].T[0]) == low


def test_snaive_prediction_interval_80_high():
    '''
    test snaive 80% upper prediction interval
    intervals are matched from R forecast package
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [32.00420, 57.00420, 49.00420, 30.00420, 26.62116]
    high = [57.99580, 82.99580, 74.99580, 55.99580, 63.37884]

    # quarterly data
    model = b.SNaive(period=4)
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.2])

    print(intervals[0].T[1])
    assert pytest.approx(intervals[0].T[1]) == high


def test_snaive_prediction_interval_95_high():
    '''
    test snaive 95% upper prediction interval
    intervals are matched from R forecast package
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [25.12464, 50.12464, 42.12464, 23.12464, 16.89199]
    high = [64.87536, 89.87536, 81.87536, 62.87536, 73.10801]

    # quarterly data
    model = b.SNaive(period=4)
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.05])

    print(intervals[0].T[1])
    assert pytest.approx(intervals[0].T[1]) == high


def test_snaive_prediction_interval_95_low():
    '''
    test snaive 95% lower prediction interval
    intervals are matched from R forecast package
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    low = [25.12464, 50.12464, 42.12464, 23.12464, 16.89199]
    # high = [64.87536, 89.87536, 81.87536, 62.87536, 73.10801]

    # quarterly data
    model = b.SNaive(period=4)
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.05])

    print(intervals[0].T[0])
    assert pytest.approx(intervals[0].T[0]) == low


def test_drift_prediction_interval_95_low():
    '''
    test drift 95% lower prediction interval
    intervals are matched from R forecast package
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    low = [22.2100359, 13.2828923, 6.2277574, 0.1124247, -5.4196405]
    # high = [63.70916, 72.55549, 79.52982, 85.56434, 91.01560]

    # quarterly data
    model = b.Drift()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.05])

    print(intervals[0].T[0])
    # not ideal due to not adjusting for drift i think,
    assert pytest.approx(intervals[0].T[0], rel=1e-6, abs=1.2) == low


def test_drift_prediction_interval_95_high():
    '''
    test drift 95% lower prediction interval
    intervals are matched from R forecast package
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [22.2100359, 13.2828923, 6.2277574, 0.1124247, -5.4196405]
    high = [63.70916, 72.55549, 79.52982, 85.56434, 91.01560]

    # quarterly data
    model = b.Drift()
    model.fit(train)
    _, intervals = model.predict(5, return_predict_int=True, alpha=[0.05])

    print(intervals[0].T[1])
    # not ideal due to not adjusting for drift i think,
    assert pytest.approx(intervals[0].T[1], rel=1e-6, abs=1.2) == high


def test_bootstrap_prediction_interval_length():
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [22.2100359, 13.2828923, 6.2277574, 0.1124247, -5.4196405]
    # high = [63.70916, 72.55549, 79.52982, 85.56434, 91.01560]

    # quarterly data
    model = b.Naive1()
    model.fit(train)
    preds = model.predict(horizon=5)

    expected = 5

    y_intervals = b.boot_prediction_intervals(preds, model.resid,
                                              horizon=expected,
                                              levels=[0.8], boots=10)

    assert expected == len(y_intervals[0])


@pytest.mark.parametrize("intervals, expected",
                         [([0.8, 0.90, 0.95], 3),
                          ([0.8, 0.90], 2),
                          ([0.8], 1),
                          ([0.7, 0.8, 0.9, 0.95], 4)
                          ])
def test_bootstrap_prediction_interval_sets_returned(intervals, expected):
    '''
    Test the number of bootstrap prediction intervals returned

    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=100)
    # low = [22.2100359, 13.2828923, 6.2277574, 0.1124247, -5.4196405]
    # high = [63.70916, 72.55549, 79.52982, 85.56434, 91.01560]

    # quarterly data
    model = b.Naive1()
    model.fit(train)
    preds = model.predict(horizon=5)
    horizon = 5
    y_intervals = b.boot_prediction_intervals(preds, model.resid,
                                              horizon=horizon,
                                              levels=intervals, boots=10)

    assert expected == len(y_intervals)


@pytest.mark.parametrize("training_length",
                         [(100),
                          (999),
                          (1000),
                          (10),
                          (20000)
                          ])
def test_drift_fitted_values_length(training_length):
    '''
    test drift .fittedvalues
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=training_length)

    model = b.Drift()
    model.fit(train)

    expected = training_length

    assert len(model.fittedvalues) == expected


@pytest.mark.parametrize("training_length",
                         [(100),
                          (999),
                          (1000),
                          (10),
                          (20000)
                          ])
def test_naive1_fitted_values_length(training_length):
    '''
    test naive1 .fittedvalues length is as expected
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=training_length)

    model = b.Naive1()
    model.fit(train)

    expected = training_length

    assert len(model.fittedvalues) == expected


@pytest.mark.parametrize("training_length",
                         [(100),
                          (999),
                          (1000),
                          (10),
                          (20000)
                          ])
def test_snaive_fitted_values_length(training_length):
    '''
    test SNaive .fittedvalues length is as expected
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=training_length)

    model = b.SNaive(7)
    model.fit(train)

    expected = training_length

    assert len(model.fittedvalues) == expected


@pytest.mark.parametrize("training_length",
                         [(100),
                          (999),
                          (1000),
                          (10),
                          (20000)
                          ])
def test_average_fitted_values_length(training_length):
    '''
    test Average forecaster .fittedvalues length is as expected
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=training_length)

    model = b.Average()
    model.fit(train)

    expected = training_length

    assert len(model.fittedvalues) == expected


@pytest.mark.parametrize("period",
                         [(7),
                          (14),
                          (24),
                          (1),
                          (12),
                          (4)
                          ])
def test_snaive_fitted_values_nan_length(period):
    '''
    test SNaive .fittedvalues has the correct number of NaNs
    i.e. = to the seasonal period.
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=200)

    model = b.SNaive(period)
    model.fit(train)

    expected = period
    n_nan = np.isnan(model.fittedvalues).sum()

    assert n_nan == expected


def test_naive1_fitted_values_nan_length():
    '''
    test Naive1 .fittedvalues has the correct number of NaNs
    i.e. = 1
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=200)

    model = b.Naive1()
    model.fit(train)

    expected = 1
    n_nan = np.isnan(model.fittedvalues).sum()

    assert n_nan == expected


def test_drift_fitted_values_nan_length():
    '''
    test Drift .fittedvalues has the correct number of NaNs
    i.e. = 1
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=200)

    model = b.Drift()
    model.fit(train)

    expected = 1
    n_nan = np.isnan(model.fittedvalues).sum()

    assert n_nan == expected


def test_average_fitted_values_nan_length():
    '''
    test Average forecast .fittedvalues has the correct number of NaNs
    i.e. = 1
    '''
    np.random.seed(1066)
    train = np.random.poisson(lam=50, size=200)

    model = b.Average()
    model.fit(train)

    expected = 0
    n_nan = np.isnan(model.fittedvalues).sum()

    assert n_nan == expected


def test_snaive_call_predict_before_fit():
    '''
    test SNaive raises correct exceptions when
    predict is called before fit
    '''
    model = b.SNaive(7)
    with pytest.raises(UnboundLocalError):
        model.predict(10)


def test_drift_call_predict_before_fit():
    '''
    test Drift raises correct exceptions when
    predict is called before fit
    '''
    model = b.Drift()
    with pytest.raises(UnboundLocalError):
        model.predict(10)
